﻿using System;
using System.Collections.Generic;
using System.Linq.Expressions;
using System.Reflection;

/// <summary>
/// Base class that provides a dynamic, memberwise Equals/GetHashCode 
/// implementation. Rather than using reflection, methods are created 
/// using compiled expression trees.
/// </summary>
public class MemberwiseEqualityObject {

    static Dictionary<Type, MemberwiseFunctions> _functions;

    /// <summary>
    /// Used to hold delegates for the compiled methods.
    /// </summary>
	private class MemberwiseFunctions {
        /// <summary>
        /// Delegate for the Equals method.
        /// </summary>
        public Func<object, object, bool> EqualsFunc;
        /// <summary>
        /// Delegate for the GetHashCode method.
        /// </summary>
        public Func<object, int> GetHashCodeFunc;
	}

    /// <summary>
    /// Static contructor.
    /// </summary>
	static MemberwiseEqualityObject() {
		_functions = new Dictionary<Type, MemberwiseFunctions>();
	}

    /// <summary>
    /// Creates the GetHashCode() method.
    /// </summary>
    /// <param name="type"></param>
    /// <returns></returns>
    static Func<object, int> MakeGetHashCodeMethod(Type type) {
        ParameterExpression pThis = Expression.Parameter(typeof(object), "x");
        UnaryExpression pCastThis = Expression.Convert(pThis, type);

        Expression last = null;
        foreach (PropertyInfo property in type.GetProperties()) {
            MethodCallExpression hash = Expression.Call(Expression.Property(pCastThis, property), "GetHashCode", Type.EmptyTypes);

            if (last == null)
                last = hash;
            else
                last = Expression.ExclusiveOr(last, hash);
        }

        return Expression.Lambda<Func<object, int>>(last, pThis).Compile();
    }

    /// <summary>
    /// Creates the Equals() method.
    /// </summary>
    /// <param name="type"></param>
    /// <returns></returns>
    static Func<object, object, bool> MakeEqualsMethod(Type type) {
        ParameterExpression pThis = Expression.Parameter(typeof(object), "x");
        ParameterExpression pThat = Expression.Parameter(typeof(object), "y");

        // cast to the subclass type
        UnaryExpression pCastThis = Expression.Convert(pThis, type);
        UnaryExpression pCastThat = Expression.Convert(pThat, type);

        // compound AND expression using short-circuit evaluation
        Expression last = null;
        foreach (PropertyInfo property in type.GetProperties()) {
            BinaryExpression equals = Expression.Equal(
                Expression.Property(pCastThis, property),
                Expression.Property(pCastThat, property)
            );

            if (last == null)
                last = equals;
            else
                last = Expression.AndAlso(last, equals);
        }

        // call Object.Equals if second parameter doesn't match type
        last = Expression.Condition(
            Expression.TypeIs(pThat, type),
            last,
            Expression.Equal(pThis, pThat)
        );

        // compile method
        return Expression.Lambda<Func<object, object, bool>>(last, pThis, pThat).Compile();
    }

    /// <summary>
    /// Dynamically compiles the Equals/GetHashCode functions on the 
    /// first call to a subclass constructor.
    /// </summary>
    public MemberwiseEqualityObject() {
        Type type = GetType();
        
        if (!_functions.ContainsKey(type)) {
            MemberwiseFunctions funcs = new MemberwiseFunctions();

            funcs.EqualsFunc = MakeEqualsMethod(type);
            funcs.GetHashCodeFunc = MakeGetHashCodeMethod(type);

            _functions[type] = funcs;
        }
    }

	/// <summary>
	/// Returns the member-wise hash code for this instance.
	/// </summary>
	/// <returns></returns>
	public override int GetHashCode() {
		return _functions[GetType()].GetHashCodeFunc(this);
	}

	/// <summary>
	/// Determines whether two instances are equal, using a member-wise comparison.
	/// </summary>
	/// <param name="obj"></param>
	/// <returns></returns>
	public override bool Equals(object obj) {
        return _functions[GetType()].EqualsFunc(this, obj);
	}
}
