﻿using System;
using System.Collections.Generic;

class Program {

    const int NUM_INSTANCES = 5000000;
    const int NUM_RUNS = 4;

    static double _timeMemberwise;
    static double _timeReflection;
    static double _timeAnonymous;

    static void Main(string[] args) {
        Random r = new Random();

        _timeMemberwise = _timeReflection = _timeAnonymous = 0;

        List<object> memberwise = new List<object>();
        List<object> reflection = new List<object>();
        List<object> anonymous = new List<object>();

        Console.Write("Creating test instances");

        // create test instances of each class from the same set of random values
        int mod = NUM_INSTANCES / 10;
        for (int i = 0; i < NUM_INSTANCES; i++) {
            decimal[] vals = new decimal[8];
            for (int j = 0; j < 8; j++) vals[j] = r.Next(100);

            memberwise.Add(new MemberwiseTest() { A = vals[0], B = vals[1], C = vals[2], D = vals[3], E = vals[4], F = vals[5], G = vals[6], H = vals[7] });
            reflection.Add(new ReflectTest() { A = vals[0], B = vals[1], C = vals[2], D = vals[3], E = vals[4], F = vals[5], G = vals[6], H = vals[7] });
            anonymous.Add(new { A = vals[0], B = vals[1], C = vals[2], D = vals[3], E = vals[4], F = vals[5], G = vals[6], H = vals[7] });

            if (i % mod == 0) Console.Write(".");
        }

        Console.WriteLine(Environment.NewLine);

        for (int i = 0; i < NUM_RUNS; i++) PerformTests(memberwise, reflection, anonymous);

        Console.WriteLine("-- Averages --");

        Console.WriteLine("MemberwiseEqualityObject:\t{0:N3} sec", _timeMemberwise / NUM_RUNS);
        Console.WriteLine("Anonymous types:\t\t{0:N3} sec", _timeAnonymous / NUM_RUNS);
        Console.WriteLine("ReflectEqualityObject:\t\t{0:N3} sec", _timeReflection / NUM_RUNS);
        Console.WriteLine();

        Console.Write("Press any key to exit...");
        Console.ReadLine();
    }

    /// <summary>
    /// Tests the performance of each type's memberwise equals method by comparing neighbouring instances.
    /// </summary>
    /// <param name="a"></param>
    /// <param name="b"></param>
    /// <param name="c"></param>
    private static void PerformTests(List<object> a, List<object> b, List<object> c) {
        bool equals;
        double secs;

        DateTime start = DateTime.Now;
        for (int i = 0; i < (c.Count - 1); i++) {
            equals = c[i].Equals(c[i + 1]);
        }
        Console.WriteLine("Anonymous types:\t\t{0:N3} sec", secs = DateTime.Now.Subtract(start).TotalSeconds);
        _timeAnonymous += secs;

        start = DateTime.Now;
        for (int i = 0; i < (a.Count - 1); i++) {
            equals = a[i].Equals(a[i + 1]);
        }
        Console.WriteLine("MemberwiseEqualityObject:\t{0:N3} sec", secs = DateTime.Now.Subtract(start).TotalSeconds);
        _timeMemberwise += secs;

        start = DateTime.Now;
        for (int i = 0; i < (b.Count - 1); i++) {
            equals = b[i].Equals(b[i + 1]);
        }
        Console.WriteLine("ReflectEqualityObject:\t\t{0:N3} sec", secs = DateTime.Now.Subtract(start).TotalSeconds);
        _timeReflection += secs;

        Console.WriteLine();
    }
}

/// <summary>
/// Test class extending MemberwiseEqualityObject
/// </summary>
public class MemberwiseTest : MemberwiseEqualityObject {

    public decimal A { get; set; }
    public decimal B { get; set; }
    public decimal C { get; set; }
    public decimal D { get; set; }
    public decimal E { get; set; }
    public decimal F { get; set; }
    public decimal G { get; set; }
    public decimal H { get; set; }
}

/// <summary>
/// Test class extending ReflectEqualityObject
/// </summary>
public class ReflectTest : ReflectEqualityObject {

    public decimal A { get; set; }
    public decimal B { get; set; }
    public decimal C { get; set; }
    public decimal D { get; set; }
    public decimal E { get; set; }
    public decimal F { get; set; }
    public decimal G { get; set; }
    public decimal H { get; set; }
}
