﻿using System;
using System.Collections.Generic;
using System.Reflection;

/// <summary>
/// Base class that provides a dynamic, memberwise Equals/GetHashCode 
/// implementation. Uses reflection.
/// </summary>
public class ReflectEqualityObject {

    static Dictionary<Type, PropertyInfo[]> _properties;

    /// <summary>
    /// Static constructor.
    /// </summary>
    static ReflectEqualityObject() {
        _properties = new Dictionary<Type, PropertyInfo[]>();
    }

    /// <summary>
    /// Caches the public properties on the first call to the subclass 
    /// constructor.
    /// </summary>
    public ReflectEqualityObject() {
        Type type = GetType();

        if (!_properties.ContainsKey(type)) {
            _properties[type] = type.GetProperties();
        }
    }

    /// <summary>
    /// Determines whether two instances are equal, using a memberwise 
    /// comparison.
    /// </summary>
    /// <param name="obj"></param>
    /// <returns></returns>
    public override bool Equals(object obj) {
        Type type = GetType();
        
        if (type.IsAssignableFrom(obj.GetType())) {
            foreach (PropertyInfo property in _properties[type]) {
                if (!property.GetValue(this, null).Equals(property.GetValue(obj, null))) return false;
            }

            return true;
        }
        else {
            return base.Equals(obj);
        }
    }

    /// <summary>
    /// Returns the member-wise hash code for this instance.
    /// </summary>
    /// <returns></returns>
    public override int GetHashCode() {
        int hash = 0;
        foreach (PropertyInfo property in _properties[GetType()]) {
            hash ^= property.GetValue(this, null).GetHashCode();
        }

        return hash;
    }
}